#! /usr/bin/env python
# -*- coding: utf_8 -*-
""" Interactive Scene provide event system, based on tags, included in 
 Panda3D model files.
 Version:   0.1.0
 Author:    Ninth 
 E-mail:    09th@nm.ru
"""
from panda3d.core import *
from direct.showbase.DirectObject import DirectObject
from direct.actor.Actor import Actor as BaseActor
from direct.interval.IntervalGlobal import *
import re

class InteractiveScene(DirectObject):
    root = None
    
    def __init__(self, events = []):
        self.elements = {}
        self.functions = {}
        self.intervals = {}
        self.possibleEvents = events
        self.accept('interactive-scene-event', self.processEvent)
        #------------- Register intervals -------------
        self.intervals['fn'] = self._intervalFunction
        self.intervals['ac'] = self._intervalActor
        self.intervals['sh'] = self._intervalShape
        self.intervals['wt'] = self._intervalWait
        #-------- Register builtin functions ----------
        self.functions['_print'] = self._builtinPrint
        self.functions['_enable'] = self._builtinEnable
        self.functions['_disable'] = self._builtinDisable
        self.functions['_stopEvt'] = self._builtinStop
        self.functions['_stopObj'] = self._builtinStopObject
        self.functions['_play'] = self._builtinPlayAnim
        self.functions['_loop'] = self._builtinLoopAnim
        self.functions['_stop'] = self._builtinStopAnim
        self.functions['_hide'] = self._builtinHideNode
        self.functions['_show'] = self._builtinShowNode
    
    def _makeDict(self, txt, globalLevel = 0, storage = {}):
        lvl = 0
        last = 0
        first = 0
        scnt = 0
        smbl = '{}'
        stype = 'parallel'
        if txt.find('{') < 0 and txt.find('[') < 0:
            #print '\t' * globalLevel, '<end>'
            storage['do'] = txt
        else:
            for c in txt:
                if lvl == 0:
                    if c == '{':
                        smbl = '{}'
                        type = 'parallel'
                    elif c == '[':
                        smbl = '[]'
                        type = 'sequence'

                if c == smbl[0]:
                    if lvl == 0:
                        first = last + 1
                        if scnt == 0:
                            storage[type] = []
                        storage[type].append({})

                    lvl += 1
                if c == smbl[1]: 
                    lvl -= 1
                    if lvl == 0:
                        #print '\t' * globalLevel, txt[first:last]
                        self._makeDict(txt[first:last], globalLevel + 1, storage[type][scnt])
                        scnt += 1
                last += 1
        return storage
        

    def _parseDict(self, dicts, lvl = 0, control = None, evtInfo = None):
        if isinstance(dicts, type([])):
            for dict in dicts:
                if isinstance(dict, type({})):
                    for key, value in dict.items():
                        #print '-'*lvl, key
                        newcontrol = None
                        if key == 'parallel':
                            newcontrol = Parallel()
                        elif key == 'sequence':
                            newcontrol = Sequence()
                        elif key == 'do':
                            control.append(self._makeInterval(value, evtInfo))
                        if key in ['parallel', 'sequence']:
                            control.append(newcontrol)
                            self._parseDict(value, lvl + 1, newcontrol, evtInfo)
        else:
            if isinstance(dicts, type({})):
                key = dicts.keys()[0]
                #print '-'*lvl, key
                newcontrol = None
                if key == 'parallel': newcontrol = Parallel()
                elif key == 'sequence': newcontrol = Sequence()
                if key in ['parallel', 'sequence']:
                    self._parseDict(dicts[key], lvl, newcontrol, evtInfo)
                return newcontrol

    def _makeInterval(self,txt, evtInfo = None):
        kwds = txt.split(':')
        if len(kwds) > 1:
            key = kwds[0].strip()
            if key in self.intervals.keys():
                return self.intervals[key](evtInfo, kwds)
            else:
                return Func(self._builtinPrint,'ERROR: incorrect keyword "' + key + '"')
            """
            if key      == 'fn':    #custom function
                fname = kwds[1]
                args = [kwds[i] for i in xrange(2,len(kwds))]
                if self.functions.has_key(fname):
                    return Func(self.functions[fname], evtInfo, *args)
                else:
                    print 'ERROR: Not found function ' + fname + '.'
            elif key    == 'ac':    #actor interval animate
                if self.root:
                    if len(kwds)>5:
                        if len(kwds)<7:
                            kwds.append(1.0)
                        if len(kwds)<8:
                            kwds.append('modelRoot')
                        return self.dynamic.actorInterval(
                            kwds[1],
                            loop=int(kwds[2]),
                            duration=float(kwds[3]),
                            #startTime=T1,
                            #endTime=T2,
                            startFrame=int(kwds[4]),
                            endFrame=int(kwds[5]),
                            playRate=float(kwds[6]),
                            partName=kwds[7]
                            #lodName=LN,
                        )
            elif key    == 'sh':    #shape key interval animate
                if self.root:
                    if 4<len(kwds)<7:
                        if len(kwds) == 6: btype = kwds[5]
                        else: btype = 'noBlend'
                        return LerpFunc(self.dynamic.shapekeyControl,
                                fromData = float(kwds[2]),
                                toData = float(kwds[3]),
                                duration = float(kwds[4]),
                                blendType = btype,
                                extraArgs = [kwds[1],],
                                name = None)

            elif key    == 'wt':    #wait function
                return Wait(float(kwds[1]))
            """

    def analyzeEvtString(self, txt, evtInfo = None):
        elist = txt.split(';')
        controls = []
        types = []
        for etxt in elist:
            dict = self._makeDict(etxt)
            controls.append(self._parseDict(dict, 0, None, evtInfo))
            tp = re.search('^[^\[{]+?[\[{]',etxt)
            if tp:
                tp = tp.group()
                tp = tp[:-1]
            else:
                tp = 'play'
            types.append(tp)
        return {'types':types,'controls':controls}
        
    def process(self,np):
        self.root = render.attachNewNode('Scene Root')
        #self.static = self.root.attachNewNode(loader.loadModel(modelpath).node())
        self.static = self.root.attachNewNode(np.node())
        self.dynamic = Actor(np)
        self.static.find('**/+Character').removeNode()        
        self.dynamic.setBlend(frameBlend = True)
        self.dynamic.reparentTo(self.root)
        self.findEvents(self.root)
        self.findSubparts()
        
    def setEvent(self,objName,evtType,evtStr):
        print 'SET EVENT', objName, evtType, evtStr
        aRes = self.analyzeEvtString(evtStr,{'object':objName,'event':evtType})
        if not self.elements.has_key(objName):
            self.elements[objName] = ActiveElement(objName, self.possibleEvents)
        self.elements[objName].evtdata[evtType].control = aRes['controls']
        self.elements[objName].evtdata[evtType].etype = aRes['types']

        
    def findEvents(self, np, lvl = 0):
        for child in np.getChildren():
            for evttype in self.possibleEvents:
                if child.hasTag(evttype):
                    tagv = child.getTag(evttype)
                    i = 0
                    while child.hasTag(evttype + str(i)):
                        tagv += child.getTag(evttype + str(i))
                        i += 1
                    tagv = tagv.replace('\n', ' ')
                    #tagv = tagv.replace('\r', '')
                    #print '    ' * lvl, 'Find event:', child.getName(), evttype, tagv
                    self.setEvent(child.getName(), evttype, tagv)
            self.findEvents(child, lvl + 1)
            
    def _iterSubpart(self,part):
        for child in part.getChildren():
            #print '+', child.getName()
            if isinstance(child, CharacterJoint): 
                #np = self.dynamic.exposeJoint(None,'modelRoot',child.getName())
                sp = re.search('^_sub_(.+?)$',child.getName())
                if sp: 
                    #print sp.groups()[0]
                    self.dynamic.makeSubpart(sp.groups()[0], [child.getName(),])
                    print 'SUBPART:', sp.groups()[0], [child.getName(),]
            self._iterSubpart(child)
    
    def findSubparts(self):
        self._iterSubpart(self.dynamic.getPartBundle('modelRoot'))
        #for i in xrange(len(self.dynamic.getPart('modelRoot').node().getBundles())):
        #    self._iterSubpart(self.dynamic.getPart('modelRoot').node().getBundle(i))
            
    def processEvent(self, objName, evtType):
        #print 'EVENT:', objName, evtType
        if self.elements.has_key(objName):
            self.elements[objName].doEvent(evtType)
    
    def _getObjectEventWrapper(self, evtInfo, objName):
        if objName == '_self':
            if evtInfo:
                if evtInfo.has_key('object'):
                    objName = evtInfo['object']
        if self.elements.has_key(objName):
            return self.elements[objName]
        else:
            return None
            
    #-------------------------------------------------------------------
    #|                   Intervals representation                      |
    #-------------------------------------------------------------------
    
    def _intervalFunction(self, evtInfo, kwds):    #custom function
        fname = kwds[1]
        args = [kwds[i] for i in xrange(2,len(kwds))]
        if self.functions.has_key(fname):
            return Func(self.functions[fname], evtInfo, *args)
        else:
            print 'ERROR: Not found function ' + fname + '.'
            return Sequence()
    
    def _intervalActor(self, evtInfo, kwds):    #actor interval animate
        if self.root:
            if len(kwds)>5:
                if len(kwds)<7:
                    kwds.append(1.0)
                if len(kwds)<8:
                    kwds.append('modelRoot')
                return self.dynamic.actorInterval(
                    kwds[1],
                    loop=int(kwds[2]),
                    duration=float(kwds[3]),
                    #startTime=T1,
                    #endTime=T2,
                    startFrame=int(kwds[4]),
                    endFrame=int(kwds[5]),
                    playRate=float(kwds[6]),
                    partName=kwds[7]
                    #lodName=LN,
                )
    
    def _intervalShape(self, evtInfo, kwds):    #shape key interval animate
        if self.root:
            if 4<len(kwds)<7:
                if len(kwds) == 6: btype = kwds[5]
                else: btype = 'noBlend'
                return LerpFunc(self.dynamic.shapekeyControl,
                        fromData = float(kwds[2]),
                        toData = float(kwds[3]),
                        duration = float(kwds[4]),
                        blendType = btype,
                        extraArgs = [kwds[1],],
                        name = None)

    def _intervalWait(self, evtInfo, kwds):    #wait function
        return Wait(float(kwds[1]))
            
    #-------------------------------------------------------------------
    #|               Builtin Accessible Functions                      |
    #-------------------------------------------------------------------
    
    def _builtinStop(self, evtInfo, objName, evt, trigger = 0):
        obj = self._getObjectEventWrapper(evtInfo, objName)
        if obj: obj.stopEvent(evt, trigger)

    def _builtinStopObject(self, evtInfo, objName):
        obj = self._getObjectEventWrapper(evtInfo, objName)
        if obj: obj.stopAll()
    
    def _builtinPrint(self, evtInfo, *args):
        txt = ' '.join(args)
        print txt
        
    def _builtinDisable(self, evtInfo, *args):
        obj = self._getObjectEventWrapper(evtInfo, args[0])
        if obj:
            for i in xrange(1,len(args)):
                obj.disableEvt(args[i])
        
    def _builtinEnable(self, evtInfo, *args):
        obj = self._getObjectEventWrapper(evtInfo, args[0])
        if obj:
            for i in xrange(1,len(args)):
                obj.enableEvt(args[i])
                
    def _builtinPlayAnim(self, evtInfo, aName, fromF, toF, pRate = 1, partN = 'modelRoot', rst = True):
        #print evtInfo, aName, fromF, toF, pRate, partN
        self.dynamic.setPlayRate(float(pRate), aName, partName = partN)
        self.dynamic.play(aName, fromFrame = int(fromF), toFrame = int(toF), partName = partN, restart = rst)
        
        
    def _builtinLoopAnim(self, evtInfo, aName, fromF, toF, pRate = 1, partN = 'modelRoot', rst = False):
        #print evtInfo, aName, fromF, toF, pRate, partN
        self.dynamic.setPlayRate(float(pRate), aName, partName = partN)
        self.dynamic.loop(aName, fromFrame = int(fromF), toFrame = int(toF), partName = partN, restart = rst)
        
    def _builtinStopAnim(self, evtInfo, aName, partN = 'modelRoot'):
        #print evtInfo, aName, partN
        self.dynamic.stop(aName, partName = partN)
        
    def _builtinHideNode(self, evtInfo, name):
        scene.root.find('**/'+name).hide()
        
    def _builtinShowNode(self, evtInfo, name):
        scene.root.find('**/'+name).show()


class Actor(BaseActor):
    shapekeys = {}
    def shapekeyControl(self, val, name):
        if not self.shapekeys.has_key(name):
            self.shapekeys[name] = self.controlJoint(None,"modelRoot",name)
        self.shapekeys[name].setX(val)


class ActiveElement(object):
    enabled = True
    name = ''
    
    def __init__(self, name, pEvtLst):
        self.name = name
        self.evtdata = {}
        for pe in pEvtLst:
            self.evtdata[pe] = EventData()
 
    
    def doEvent(self,etype):
        if self.evtdata.has_key(etype):
            data = self.evtdata[etype]
            if data.enabled:
                if data.control:
                    if data.etype[data.current] == 'loop':
                        if data.control[data.current].isStopped():
                            data.control[data.current].loop()
                        else:
                            data.control[data.current].resume()
                        if len(data.control) > 1 and not data.firstrun:
                            old = data.current - 1
                            if old < 0: old = len(data.control) - 1
                            data.control[old].pause()
                    elif data.etype[data.current] == 'play':
                        data.control[data.current].start()
                        if len(data.control) > 1 and not data.firstrun:
                            old = data.current - 1
                            if old < 0: old = len(data.control) - 1
                            data.control[old].finish()
                    if data.etype[data.current] in ['loop','play']:
                        data.current += 1
                        data.firstrun = False
                        if data.current >= len(data.control):
                            data.current = 0
        else:
            print 'WARNING (' + self.name + '): Event ' + etype + ' not found.'
            
            
    def stopEvent(self, evt, trigger):
        if self.evtdata.has_key(evt):
            data = self.evtdata[evt]
            if len(data.control) > trigger:
                if data.etype[trigger] == 'play':
                    data.control[trigger].finish()
                elif data.etype[trigger] == 'loop':
                    data.control[trigger].pause()
        
    def stopAll(self):
        for key, data in self.evtdata.items():
            if data.control:
                for i in xrange(len(data.control)):
                    if data.etype[i] == 'play':
                        data.control[i].finish()
                    elif data.etype[i] == 'loop':
                        data.control[i].pause()
    
    def disableEvt(self,evt):
        if self.evtdata.has_key(evt):
            self.evtdata[evt].enabled = False
            
    def enableEvt(self,evt):
        if self.evtdata.has_key(evt):
            self.evtdata[evt].enabled = True

class EventData:
    control = None
    current = 0
    etype = None
    enabled = True
    firstrun = True
